@@ -1,9 +1,8 @@
-import os
-import sys
 import inspect
+import sys
 import traceback
-from functools import partial, wraps
 from datetime import datetime
+from functools import wraps
 
 import click
 
@@ -13,21 +12,27 @@ from . import parameters
 from . import decorators
 from . import metaflow_version
 from . import namespace
-from .util import resolve_identity, decompress_list, write_latest_run_id, get_latest_run_id
+from . import current
+from .cli_args import cli_args
+from .util import resolve_identity, decompress_list, write_latest_run_id, \
+    get_latest_run_id, to_unicode
 from .task import MetaflowTask
 from .exception import CommandException, MetaflowException
 from .graph import FlowGraph
 from .datastore import DATASTORES
 from .runtime import NativeRuntime
 from .package import MetaflowPackage
-from .plugins import LOGGING_SIDECAR, MONITOR_SIDECAR
-from .metadata import METADATAPROVIDERS
-from .metaflow_config import DEFAULT_DATASTORE, DEFAULT_METADATA
-from .plugins import ENVIRONMENTS
-from .environment import MetaflowEnvironment
+from .plugins import ENVIRONMENTS, LOGGING_SIDECARS, METADATA_PROVIDERS, MONITOR_SIDECARS
+from .metaflow_config import DEFAULT_DATASTORE, DEFAULT_ENVIRONMENT, DEFAULT_EVENT_LOGGER, \
+    DEFAULT_METADATA, DEFAULT_MONITOR, DEFAULT_PACKAGE_SUFFIXES
+from .metaflow_environment import MetaflowEnvironment
 from .pylint_wrapper import PyLint
 from .event_logger import EventLogger
 from .monitor import Monitor
+from .R import use_r, metaflow_r_version
+from .mflog import mflog, LOG_SOURCES
+from .unbounded_foreach import UBF_CONTROL, UBF_TASK
+
 
 ERASE_TO_EOL = '\033[K'
 HIGHLIGHT = 'red'
@@ -89,9 +94,17 @@ def echo_always(line, **kwargs):
         click.secho(ERASE_TO_EOL, **kwargs)
 
 
-def logger(body='', system_msg=False, head='', bad=False, timestamp=True):
+def logger(body='',
+           system_msg=False,
+           head='',
+           bad=False,
+           timestamp=True):
     if timestamp:
-        tstamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
+        if timestamp is True:
+            dt = datetime.now()
+        else:
+            dt = timestamp
+        tstamp = dt.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
         click.secho(tstamp + ' ', fg=LOGGER_TIMESTAMP, nl=False)
     if head:
         click.secho(head, fg=LOGGER_COLOR, nl=False)
@@ -269,8 +282,17 @@ def dump(obj,
               default=True,
               show_default=True,
               help='Show both stdout and stderr of the task.')
+@click.option('--timestamps/--no-timestamps',
+              default=False,
+              show_default=True,
+              help='Show timestamps.')
 @click.pass_obj
-def logs(obj, input_path, stdout=None, stderr=None, both=None):
+def logs(obj,
+         input_path,
+         stdout=None,
+         stderr=None,
+         both=None,
+         timestamps=False):
     types = set()
     if stdout:
         types.add('stdout')
@@ -281,6 +303,8 @@ def logs(obj, input_path, stdout=None, stderr=None, both=None):
     if both:
         types.update(('stdout', 'stderr'))
 
+    streams = list(sorted(types, reverse=True))
+
     # Pathspec can either be run_id/step_name or run_id/step_name/task_id.
     parts = input_path.split('/')
     if len(parts) == 2:
@@ -289,7 +313,7 @@ def logs(obj, input_path, stdout=None, stderr=None, both=None):
     elif len(parts) == 3:
         run_id, step_name, task_id = parts
     else:
-        raise CommandException("input_path should either be run_id/step_name"
+        raise CommandException("input_path should either be run_id/step_name "
                                "or run_id/step_name/task_id")
 
     if obj.datastore.datastore_root is None:
@@ -300,31 +324,71 @@ def logs(obj, input_path, stdout=None, stderr=None, both=None):
             "Could not find the location of the datastore -- did you correctly set the "
             "METAFLOW_DATASTORE_SYSROOT_%s environment variable" % (obj.datastore.TYPE).upper())
 
-    from metaflow.datastore.datastore_set import MetaflowDatastoreSet
-    datastore_set = MetaflowDatastoreSet(
-                        obj.datastore,
-                        obj.flow.name,
-                        run_id,
-                        steps=[step_name],
-                        metadata=obj.metadata,
-                        monitor=obj.monitor,
-                        event_logger=obj.event_logger)
     if task_id:
-        ds_list = [datastore_set.get_with_pathspec(input_path)]
+        ds_list = [obj.datastore(obj.flow.name,
+                                 run_id=run_id,
+                                 step_name=step_name,
+                                 task_id=task_id,
+                                 mode='r',
+                                 allow_unsuccessful=True)]
     else:
-        ds_list = list(datastore_set) # get all tasks
-    for ds in ds_list:
-        echo('Dumping logs of run_id=*{run_id}* '
-             'step=*{step}* task_id=*{task_id}*'.format(run_id=ds.run_id,
-                                                        step=ds.step_name,
-                                                        task_id=ds.task_id),
-             fg='magenta')
+        from metaflow.datastore.datastore_set import MetaflowDatastoreSet
+        datastore_set = MetaflowDatastoreSet(
+                            obj.datastore,
+                            obj.flow.name,
+                            run_id,
+                            steps=[step_name],
+                            metadata=obj.metadata,
+                            monitor=obj.monitor,
+                            event_logger=obj.event_logger)
+        # get all successful tasks
+        ds_list = list(datastore_set)
 
-        for typ in ('stdout', 'stderr'):
-            if typ in types:
-                echo(typ, bold=True)
-                click.secho(ds.load_log(typ).decode('UTF-8', errors='replace'),
-                            nl=False)
+    if ds_list:
+        def echo_unicode(line, **kwargs):
+            click.secho(line.decode('UTF-8', errors='replace'), **kwargs)
+
+        # old style logs are non mflog-style logs
+        maybe_old_style = True
+        for ds in ds_list:
+            echo('Dumping logs of run_id=*{run_id}* '
+                 'step=*{step}* task_id=*{task_id}*'.format(run_id=ds.run_id,
+                                                            step=ds.step_name,
+                                                            task_id=ds.task_id),
+                 fg='magenta')
+
+            for stream in streams:
+                echo(stream, bold=True)
+                logs = ds.load_logs(LOG_SOURCES, stream)
+                if any(data for _, data in logs):
+                    # attempt to read new, mflog-style logs
+                    for line in mflog.merge_logs([blob for _, blob in logs]):
+                        if timestamps:
+                            ts = mflog.utc_to_local(line.utc_tstamp)
+                            tstamp = ts.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
+                            click.secho(tstamp + ' ',
+                                        fg=LOGGER_TIMESTAMP,
+                                        nl=False)
+                        echo_unicode(line.msg)
+                    maybe_old_style = False
+                elif maybe_old_style:
+                    # if they are not available, we may be looking at
+                    # a legacy run (unless we have seen new-style data already
+                    # for another stream). This return an empty string if
+                    # nothing is found
+                    log = ds.load_log_legacy(stream)
+                    if log and timestamps:
+                        raise CommandException("We can't show --timestamps for "
+                                               "old runs. Sorry!")
+                    echo_unicode(log, nl=False)
+
+    elif len(parts) == 2:
+        # TODO if datastore provided a way to find unsuccessful task IDs, we
+        # could make handle this case automatically
+        raise CommandException("Successful tasks were not found at the given "
+                               "path. You can see logs for unsuccessful tasks "
+                               "by giving an exact task ID using the "
+                               "run_id/step_name/task_id format.")
 
 
 # TODO - move step and init under a separate 'internal' subcommand
@@ -349,14 +413,14 @@ def logs(obj, input_path, stdout=None, stderr=None, both=None):
               show_default=True,
               help='Index of this foreach split.')
 @click.option('--tag',
-              'tags',
+              'opt_tag',
               multiple=True,
               default=None,
               help="Annotate this run with the given tag. You can specify "
                    "this option multiple times to attach multiple tags in "
                    "the task.")
 @click.option('--namespace',
-              'user_namespace',
+              'opt_namespace',
               default=None,
               help="Change namespace from the default (your username) to "
                    "the specified tag.")
@@ -374,25 +438,38 @@ def logs(obj, input_path, stdout=None, stderr=None, both=None):
               default=None,
               help="Run id of the origin flow, if this task is part of a flow "
               "being resumed.")
-@click.pass_obj
-def step(obj,
+@click.option('--with',
+              'decospecs',
+              multiple=True,
+              help="Add a decorator to this task. You can specify this "
+              "option multiple times to attach multiple decorators "
+              "to this task.")
+@click.option('--ubf-context',
+              default='none',
+              type=click.Choice(['none', UBF_CONTROL, UBF_TASK]),
+              help="Provides additional context if this task is of type "
+              "unbounded foreach.")
+@click.pass_context
+def step(ctx,
          step_name,
-         tags=None,
+         opt_tag=None,
          run_id=None,
          task_id=None,
          input_paths=None,
          split_index=None,
-         user_namespace=None,
+         opt_namespace=None,
          retry_count=None,
          max_user_code_retries=None,
          clone_only=None,
-         clone_run_id=None):
-    if user_namespace is not None:
-        namespace(user_namespace or None)
+         clone_run_id=None,
+         decospecs=None,
+         ubf_context=None):
+    if opt_namespace is not None:
+        namespace(opt_namespace or None)
 
     func = None
     try:
-        func = getattr(obj.flow, step_name)
+        func = getattr(ctx.obj.flow, step_name)
     except:
         raise CommandException("Step *%s* doesn't exist." % step_name)
     if not func.is_step:
@@ -401,20 +478,33 @@ def step(obj,
          fg='magenta',
          bold=False)
 
-    obj.datastore.datastore_root = obj.datastore_root
-    if obj.datastore.datastore_root is None:
-        obj.datastore.datastore_root = obj.datastore.get_datastore_root_from_config(obj.echo)
+    if decospecs:
+        decorators._attach_decorators_to_step(func, decospecs)
 
-    obj.metadata.add_sticky_tags(tags=tags)
+    ctx.obj.datastore.datastore_root = ctx.obj.datastore_root
+    if ctx.obj.datastore.datastore_root is None:
+        ctx.obj.datastore.datastore_root = \
+            ctx.obj.datastore.get_datastore_root_from_config(ctx.obj.echo)
+
+    step_kwargs = ctx.params
+    # Remove argument `step_name` from `step_kwargs`.
+    step_kwargs.pop('step_name', None)
+    # Remove `opt_*` prefix from (some) option keys.
+    step_kwargs = dict([(k[4:], v) if k.startswith('opt_') else (k, v)
+                         for k, v in step_kwargs.items()])
+    cli_args._set_step_kwargs(step_kwargs)
+
+    ctx.obj.metadata.add_sticky_tags(tags=opt_tag)
     paths = decompress_list(input_paths) if input_paths else []
 
-    task = MetaflowTask(obj.flow,
-                        obj.datastore,
-                        obj.metadata,
-                        obj.environment,
-                        obj.logger,
-                        obj.event_logger,
-                        obj.monitor)
+    task = MetaflowTask(ctx.obj.flow,
+                        ctx.obj.datastore,
+                        ctx.obj.metadata,
+                        ctx.obj.environment,
+                        ctx.obj.echo,
+                        ctx.obj.event_logger,
+                        ctx.obj.monitor,
+                        ubf_context)
     if clone_only:
         task.clone_only(step_name,
                         run_id,
@@ -432,7 +522,7 @@ def step(obj,
 
     echo('Success', fg='green', bold=True, indent=True)
 
-@parameters.add_custom_parameters
+@parameters.add_custom_parameters(deploy_mode=False)
 @cli.command(help="Internal command to initialize a run.")
 @click.option('--run-id',
               default=None,
@@ -442,8 +532,13 @@ def step(obj,
               default=None,
               required=True,
               help='ID for this instance of the step.')
+@click.option('--tag',
+              'tags',
+              multiple=True,
+              default=None,
+              help="Tags for this instance of the step.")
 @click.pass_obj
-def init(obj, run_id=None, task_id=None, **kwargs):
+def init(obj, run_id=None, task_id=None, tags=None, **kwargs):
     # init is a separate command instead of an option in 'step'
     # since we need to capture user-specified parameters with
     # @add_custom_parameters. Adding custom parameters to 'step'
@@ -453,7 +548,10 @@ def init(obj, run_id=None, task_id=None, **kwargs):
     # variables.
 
     if obj.datastore.datastore_root is None:
-        obj.datastore.datastore_root = obj.datastore.get_datastore_root_from_config(obj.echo)
+        obj.datastore.datastore_root = \
+            obj.datastore.get_datastore_root_from_config(obj.echo)
+
+    obj.metadata.add_sticky_tags(tags=tags)
 
     runtime = NativeRuntime(obj.flow,
                             obj.graph,
@@ -539,6 +637,14 @@ def resume(obj,
     if step_to_rerun is None:
         clone_steps = set()
     else:
+        # validate step name
+        if step_to_rerun not in obj.graph.nodes:
+            raise CommandException(
+                "invalid step name {0} specified, must be step present in "
+                "current form of execution graph. Valid step names include: {1}"
+                .format(
+                    step_to_rerun,
+                    ",".join(list(obj.graph.nodes.keys()))))
         clone_steps = {step_to_rerun}
 
     runtime = NativeRuntime(obj.flow,
@@ -562,7 +668,7 @@ def resume(obj,
     write_run_id(run_id_file, runtime.run_id)
 
 
-@parameters.add_custom_parameters
+@parameters.add_custom_parameters(deploy_mode=True)
 @cli.command(help='Run the workflow locally.')
 @common_run_options
 @click.option('--namespace',
@@ -585,7 +691,7 @@ def run(obj,
         user_namespace=None,
         **kwargs):
 
-    if namespace is not None:
+    if user_namespace is not None:
         namespace(user_namespace or None)
     before_run(obj, tags, decospecs + obj.environment.decospecs())
 
@@ -625,7 +731,7 @@ def before_run(obj, tags, decospecs):
     #
     # A downside is that we need to have the following decorators handling
     # in two places in this module and we need to make sure that
-    # _init_decorators doesn't get called twice.
+    # _init_step_decorators doesn't get called twice.
     if decospecs:
         decorators._attach_decorators(obj.flow, decospecs)
         obj.graph = FlowGraph(obj.flow.__class__)
@@ -633,17 +739,20 @@ def before_run(obj, tags, decospecs):
     #obj.environment.init_environment(obj.logger)
 
     if obj.datastore.datastore_root is None:
-        obj.datastore.datastore_root = obj.datastore.get_datastore_root_from_config(obj.echo)
+        obj.datastore.datastore_root = \
+            obj.datastore.get_datastore_root_from_config(obj.echo)
 
-    decorators._init_decorators(
+    decorators._init_step_decorators(
         obj.flow, obj.graph, obj.environment, obj.datastore, obj.logger)
     obj.metadata.add_sticky_tags(tags=tags)
 
     # Package working directory only once per run.
     # We explicitly avoid doing this in `start` since it is invoked for every
     # step in the run.
-    # TODO(crk): Capture time taken to package and log to keystone.
-    obj.package = MetaflowPackage(obj.flow, obj.environment, obj.logger, obj.package_suffixes)
+    obj.package = MetaflowPackage(obj.flow,
+                                  obj.environment,
+                                  obj.echo,
+                                  obj.package_suffixes)
 
 
 @cli.command(help='Print the Metaflow version')
@@ -651,7 +760,7 @@ def before_run(obj, tags, decospecs):
 def version(obj):
     echo_always(obj.version)
 
-
+@decorators.add_decorator_options
 @click.command(cls=click.CommandCollection,
                sources=[cli] + plugins.get_plugin_cli(),
                invoke_without_command=True)
@@ -662,10 +771,10 @@ def version(obj):
 @click.option('--metadata',
               default=DEFAULT_METADATA,
               show_default=True,
-              type=click.Choice([m.TYPE for m in METADATAPROVIDERS]),
+              type=click.Choice([m.TYPE for m in METADATA_PROVIDERS]),
               help='Metadata service type')
 @click.option('--environment',
-              default='local',
+              default=DEFAULT_ENVIRONMENT,
               show_default=True,
               type=click.Choice(['local'] + [m.TYPE for m in ENVIRONMENTS]),
               help='Execution environment type')
@@ -679,7 +788,7 @@ def version(obj):
 @click.option('--package-suffixes',
               help='A comma-separated list of file suffixes to include '
                    'in the code package.',
-              default='.py',
+              default=DEFAULT_PACKAGE_SUFFIXES,
               show_default=True)
 @click.option('--with',
               'decospecs',
@@ -696,14 +805,14 @@ def version(obj):
               show_default=True,
               help='Measure code coverage using coverage.py.')
 @click.option('--event-logger',
-              default='nullSidecarLogger',
+              default=DEFAULT_EVENT_LOGGER,
               show_default=True,
-              type=click.Choice(LOGGING_SIDECAR),
+              type=click.Choice(LOGGING_SIDECARS),
               help='type of event logger used')
 @click.option('--monitor',
-              default='nullSidecarMonitor',
+              default=DEFAULT_MONITOR,
               show_default=True,
-              type=click.Choice(MONITOR_SIDECAR),
+              type=click.Choice(MONITOR_SIDECARS),
               help='Monitoring backend type')
 @click.pass_context
 def start(ctx,
@@ -717,7 +826,8 @@ def start(ctx,
           pylint=None,
           coverage=None,
           event_logger=None,
-          monitor=None):
+          monitor=None,
+          **deco_options):
     global echo
     if quiet:
         echo = echo_dev_null
@@ -725,14 +835,14 @@ def start(ctx,
         echo = echo_always
 
     ctx.obj.version = metaflow_version.get_version()
+    version = ctx.obj.version
+    if use_r():
+        version = metaflow_r_version()
 
-    echo('Metaflow %s' % ctx.obj.version, fg='magenta', bold=True, nl=False)
+    echo('Metaflow %s' % version, fg='magenta', bold=True, nl=False)
     echo(" executing *%s*" % ctx.obj.flow.name, fg='magenta', nl=False)
     echo(" for *%s*" % resolve_identity(), fg='magenta')
 
-    if decospecs:
-        decorators._attach_decorators(ctx.obj.flow, decospecs)
-
     if coverage:
         from coverage import Coverage
         cov = Coverage(data_suffix=True,
@@ -741,6 +851,7 @@ def start(ctx,
                        branch=True)
         cov.start()
 
+    cli_args._set_top_kwargs(ctx.params)
     ctx.obj.echo = echo
     ctx.obj.echo_always = echo_always
     ctx.obj.graph = FlowGraph(ctx.obj.flow.__class__)
@@ -757,22 +868,49 @@ def start(ctx,
                            if e.TYPE == environment][0](ctx.obj.flow)
     ctx.obj.environment.validate_environment(echo)
 
+    ctx.obj.datastore = DATASTORES[datastore]
+    ctx.obj.datastore_root = datastore_root
+
+    # It is important to initialize flow decorators early as some of the
+    # things they provide may be used by some of the objects initialize after.
+    decorators._init_flow_decorators(ctx.obj.flow,
+                                     ctx.obj.graph,
+                                     ctx.obj.environment,
+                                     ctx.obj.datastore,
+                                     ctx.obj.logger,
+                                     echo,
+                                     deco_options)
+
     ctx.obj.monitor = Monitor(monitor, ctx.obj.environment, ctx.obj.flow.name)
     ctx.obj.monitor.start()
 
-    ctx.obj.metadata = [m for m in METADATAPROVIDERS
+    ctx.obj.metadata = [m for m in METADATA_PROVIDERS
                         if m.TYPE == metadata][0](ctx.obj.environment,
                                                   ctx.obj.flow,
                                                   ctx.obj.event_logger,
                                                   ctx.obj.monitor)
     ctx.obj.datastore = DATASTORES[datastore]
-    ctx.obj.datastore_root = datastore_root
+
+    if datastore_root is None:
+        datastore_root = \
+          ctx.obj.datastore.get_datastore_root_from_config(ctx.obj.echo)
+    ctx.obj.datastore_root = ctx.obj.datastore.datastore_root = datastore_root
+
+    if decospecs:
+        decorators._attach_decorators(ctx.obj.flow, decospecs)
+
+    # initialize current and parameter context for deploy-time parameters
+    current._set_env(flow_name=ctx.obj.flow.name, is_running=False)
+    parameters.set_parameter_context(ctx.obj.flow.name,
+                                        ctx.obj.echo,
+                                        ctx.obj.datastore)
+
     if ctx.invoked_subcommand not in ('run', 'resume'):
         # run/resume are special cases because they can add more decorators with --with,
         # so they have to take care of themselves.
         decorators._attach_decorators(
             ctx.obj.flow, ctx.obj.environment.decospecs())
-        decorators._init_decorators(
+        decorators._init_step_decorators(
             ctx.obj.flow, ctx.obj.graph, ctx.obj.environment, ctx.obj.datastore, ctx.obj.logger)
         #TODO (savin): Enable lazy instantiation of package
         ctx.obj.package = None
@@ -805,11 +943,26 @@ def _check(graph, flow, environment, pylint=True, warnings=False, **kwargs):
         fname = inspect.getfile(flow.__class__)
         pylint = PyLint(fname)
         if pylint.has_pylint():
-            pylint.run(warnings=warnings, pylint_config=environment.pylint_config(), logger=echo_always)
-            echo('Pylint is happy!',
-                 fg='green',
-                 bold=True,
-                 indent=True)
+            pylint_is_happy, pylint_exception_msg = pylint.run(
+                        warnings=warnings,
+                        pylint_config=environment.pylint_config(),
+                        logger=echo_always)
+
+            if pylint_is_happy:
+                echo('Pylint is happy!',
+                     fg='green',
+                     bold=True,
+                     indent=True)
+            else:
+                echo("Pylint couldn't analyze your code.\n\tPylint exception: %s"
+                     % pylint_exception_msg,
+                     fg='red',
+                     bold=True,
+                     indent=True)
+                echo("Skipping Pylint checks.",
+                     fg='red',
+                     bold=True,
+                     indent=True)
         else:
             echo("Pylint not found, so extra checks are disabled.",
                  fg='green',
@@ -847,8 +1000,6 @@ def main(flow, args=None, handle_exceptions=True, entrypoint=None):
     state = CliState(flow)
     state.entrypoint = entrypoint
 
-    parameters.set_parameter_context(flow.name)
-
     try:
         if args is None:
             start(auto_envvar_prefix='METAFLOW', obj=state)
