@@ -4,6 +4,7 @@ import time
 from collections import namedtuple
 from datetime import datetime
 
+from metaflow.current import current
 from metaflow.exception import MetaflowInternalError
 from metaflow.util import get_username, resolve_identity
 
@@ -12,7 +13,7 @@ DataArtifact = namedtuple('DataArtifact',
                           'name ds_type url type sha')
 
 MetaDatum = namedtuple('MetaDatum',
-                       'field value type')
+                       'field value type tags')
 
 
 class MetadataProviderMeta(type):
@@ -84,6 +85,17 @@ class MetadataProvider(object):
         '''
         return ''
 
+    def version(self):
+        '''
+        Returns the version of this provider
+
+        Returns
+        -------
+        str
+            Version of the provider
+        '''
+        return ''
+
     def new_run_id(self, tags=[], sys_tags=[]):
         '''
         Creates an ID and registers this new run.
@@ -176,7 +188,8 @@ class MetadataProvider(object):
         dict[string] -> string
             Environment variables from this metadata provider
         '''
-        raise NotImplementedError()
+        return {'METAFLOW_RUNTIME_NAME': runtime_name,
+                'USER': get_username()}
 
     def register_data_artifacts(self,
                                 run_id,
@@ -227,6 +240,15 @@ class MetadataProvider(object):
         '''
         raise NotImplementedError()
 
+    def start_task_heartbeat(self, flow_id, run_id, step_name, task_id):
+        pass
+
+    def start_run_heartbeat(self, flow_id, run_id):
+        pass
+
+    def stop_heartbeat(self):
+        pass
+
     @classmethod
     def _get_object_internal(cls, obj_type, obj_order, sub_type, sub_order, filters=None, *args):
         '''
@@ -239,7 +261,7 @@ class MetadataProvider(object):
         obj_type : string
             One of 'root', 'flow', 'run', 'step', 'task', 'artifact'
         obj_order: int
-            Order in the last ['root', 'flow', 'run', 'step', 'task', 'artifact']
+            Order in the list ['root', 'flow', 'run', 'step', 'task', 'artifact']
         sub_type : string
             Same as obj_type with the addition of 'metadata', 'self'
         sub_order:
@@ -353,8 +375,13 @@ class MetadataProvider(object):
             'system_tags': sys_tags,
             'ts_epoch': int(round(time.time() * 1000))}
 
-    def _flow_to_json(self, tags=[], sys_tags=[]):
-        return self._all_obj_elements(tags, sys_tags)
+    def _flow_to_json(self):
+        # No need to store tags, sys_tags or username at the flow level
+        # since runs are the top level logical concept, which is where we
+        # store tags, sys_tags and username
+        return {
+            'flow_id': self._flow_name,
+            'ts_epoch': int(round(time.time() * 1000))}
 
     def _run_to_json(self, run_id=None, tags=[], sys_tags=[]):
         if run_id is not None:
@@ -388,7 +415,7 @@ class MetadataProvider(object):
             return self._step_to_json(run_id, step_name, tags, sys_tags)
         if obj_type == 'run':
             return self._run_to_json(run_id, tags, sys_tags)
-        return self._flow_to_json(tags, sys_tags)
+        return self._flow_to_json()
 
     def _artifacts_to_json(self, run_id, step_name, task_id, attempt_id, artifacts):
         result = []
@@ -418,6 +445,7 @@ class MetadataProvider(object):
             'field_name': datum.field,
             'type': datum.type,
             'value': datum.value,
+            'tags': datum.tags,
             'user_name': user,
             'ts_epoch': int(round(time.time() * 1000))} for datum in metadata]
 
@@ -430,6 +458,13 @@ class MetadataProvider(object):
             'date:' + datetime.utcnow().strftime('%Y-%m-%d')]
         if env['metaflow_version']:
             tags.append('metaflow_version:' + env['metaflow_version'])
+        if 'metaflow_r_version' in env:
+            tags.append('metaflow_r_version:' + env['metaflow_r_version'])
+        if 'r_version_code' in env:
+            tags.append('r_version:' + env['r_version_code'])
+        if 'project_name' in current:
+            tags.append('project:' + current.project_name)
+            tags.append('project_branch:' + current.branch_name)
         return tags
 
     def _register_code_package_metadata(self, run_id, step_name, task_id):
@@ -441,7 +476,8 @@ class MetadataProvider(object):
             metadata.append(MetaDatum(
                 field='code-package',
                 value=json.dumps({'ds_type': code_ds, 'sha': code_sha, 'location': code_url}),
-                type='code-package'))
+                type='code-package',
+                tags=[]))
         if metadata:
             self.register_metadata(run_id, step_name, task_id, metadata)
 
@@ -477,5 +513,5 @@ class MetadataProvider(object):
         self._monitor = monitor
         self._environment = environment
         self._runtime = os.environ.get(
-            'METAFLOW_MLI_RUNTIME_NAME', 'dev')
+            'METAFLOW_RUNTIME_NAME', 'dev')
         self.add_sticky_tags(sys_tags=self._tags())
