@@ -9,9 +9,11 @@ from .datastore import Inputs, MetaflowDatastoreSet
 from .exception import MetaflowInternalError,\
     MetaflowDataMissing,\
     MetaflowExceptionWrapper
+from .unbounded_foreach import UBF_CONTROL
 from .util import all_equal,\
     get_username,\
-    resolve_identity
+    resolve_identity, \
+    unicode_type
 from .current import current
 
 from collections import namedtuple
@@ -32,7 +34,8 @@ class MetaflowTask(object):
                  environment,
                  console_logger,
                  event_logger,
-                 monitor):
+                 monitor,
+                 ubf_context):
         self.flow = flow
         self.datastore = datastore
         self.metadata = metadata
@@ -40,9 +43,10 @@ class MetaflowTask(object):
         self.console_logger = console_logger
         self.event_logger = event_logger
         self.monitor = monitor
+        self.ubf_context = ubf_context
 
     def _exec_step_function(self, step_function, input_obj=None):
-        self.environment.validate_environment(logger=self.console_logger)
+        self.environment.validate_environment(echo=self.console_logger)
         if input_obj is None:
             step_function()
         else:
@@ -54,8 +58,9 @@ class MetaflowTask(object):
         for var, param in self.flow._get_parameters():
             # make the parameter a read-only property
             # note x=x binds the current value of x to the closure
-            def property_setter(_, cls=self.flow.__class__, var=var, parameter_ds=parameter_ds):
-                v = parameter_ds[var]
+            def property_setter(
+                    _, cls=self.flow.__class__, param=param, var=var, parameter_ds=parameter_ds):
+                v = param.load_parameter(parameter_ds[var])
                 setattr(cls, var, property(fget=lambda _, val=v: val))
                 return v
 
@@ -161,7 +166,7 @@ class MetaflowTask(object):
                 # something strange happened upstream, the inputs list
                 # may not contain all inputs which should raise an exception
                 stack = inp['_foreach_stack']
-                if len(inputs) != stack[-1].num_splits:
+                if stack[-1].num_splits and len(inputs) != stack[-1].num_splits:
                     raise MetaflowDataMissing("Foreach join *%s* expected %d "
                                               "splits but only %d inputs were "
                                               "found" % (step_name,
@@ -180,7 +185,7 @@ class MetaflowTask(object):
                                             "although it follows a split step."
                                             % step_name)
 
-            if split_index is None:
+            if self.ubf_context != UBF_CONTROL and split_index is None:
                 raise MetaflowInternalError("Step *%s* follows a split step "
                                             "but no split_index is "
                                             "specified." % step_name)
@@ -214,6 +219,7 @@ class MetaflowTask(object):
                                 attempt=0,
                                 event_logger=self.event_logger,
                                 monitor=self.monitor)
+        output.init_task()
         origin_run_id, origin_step_name, origin_task_id =\
             clone_origin_task.split('/')
         # 2. initialize origin datastore
@@ -227,6 +233,30 @@ class MetaflowTask(object):
         output.clone(origin)
         output.done()
 
+    def _finalize_control_task(self):
+        # Update `_transition` which is expected by the NativeRuntime.
+        step_name = self.flow._current_step
+        next_steps = self.flow._graph[step_name].out_funcs
+        self.flow._transition = (next_steps, None, None)
+        if self.flow._task_ok:
+            # Throw an error if `_control_mapper_tasks` isn't populated.
+            mapper_tasks = self.flow._control_mapper_tasks
+            if not mapper_tasks:
+                msg = "Step *{step}* has a control task which didn't "\
+                      "specify the artifact *_control_mapper_tasks* for "\
+                      "the subsequent *{join}* step."
+                raise MetaflowInternalError(msg.format(step=step_name,
+                                                       join=next_steps[0]))
+            elif not (isinstance(mapper_tasks, list) and\
+                      isinstance(mapper_tasks[0], unicode_type)):
+                msg = "Step *{step}* has a control task which didn't "\
+                      "specify the artifact *_control_mapper_tasks* as a "\
+                      "list of strings but instead specified it as {typ} "\
+                      "with elements of {elem_typ}."
+                raise MetaflowInternalError(msg.format(step=step_name,
+                                                       typ=type(mapper_tasks),
+                                                       elem_typ=type(mapper_tasks[0])))
+
     def run_step(self,
                  step_name,
                  run_id,
@@ -256,10 +286,20 @@ class MetaflowTask(object):
                                         task_id,
                                         [MetaDatum(field='attempt',
                                                    value=str(retry_count),
-                                                   type='attempt'),
+                                                   type='attempt',
+                                                   tags=[]),
                                          MetaDatum(field='origin-run-id',
                                                    value=str(origin_run_id),
-                                                   type='origin-run-id')])
+                                                   type='origin-run-id',
+                                                   tags=[]),
+                                         MetaDatum(field='ds-type',
+                                                   value=self.datastore.TYPE,
+                                                   type='ds-type',
+                                                   tags=[]),
+                                         MetaDatum(field='ds-root',
+                                                   value=self.datastore.datastore_root,
+                                                   type='ds-root',
+                                                   tags=[])])
 
         step_func = getattr(self.flow, step_name)
         node = self.flow._graph[step_name]
@@ -277,8 +317,14 @@ class MetaflowTask(object):
                                 attempt=retry_count,
                                 event_logger=self.event_logger,
                                 monitor=self.monitor)
+        output.init_task()
 
         if input_paths:
+            control_paths = [path for path in input_paths
+                             if path.split('/')[-1].startswith('control-')]
+            if control_paths:
+                [control_path] = control_paths
+                input_paths.remove(control_path)
             # 2. initialize input datastores
             inputs = self._init_data(run_id, join_type, input_paths)
 
@@ -286,13 +332,15 @@ class MetaflowTask(object):
             self._init_foreach(step_name, join_type, inputs, split_index)
 
         # 4. initialize the current singleton
-        current._set_env(self.flow.name,
-                         run_id,
-                         step_name,
-                         task_id,
-                         origin_run_id,
-                         resolve_identity(),
-                         get_username())
+        current._set_env(flow_name=self.flow.name,
+                         run_id=run_id,
+                         step_name=step_name,
+                         task_id=task_id,
+                         retry_count=retry_count,
+                         origin_run_id=origin_run_id,
+                         namespace=resolve_identity(),
+                         username=get_username(),
+                         is_running=True)
 
         # 5. run task
         output.save_metadata('task_begin', {
@@ -303,6 +351,8 @@ class MetaflowTask(object):
         })
         logger = self.event_logger
         start = time.time()
+        self.metadata.start_task_heartbeat(self.flow.name, run_id, step_name,
+                                           task_id)
         try:
             # init side cars
             logger.start()
@@ -336,7 +386,9 @@ class MetaflowTask(object):
                                    self.flow,
                                    self.flow._graph,
                                    retry_count,
-                                   max_user_code_retries)
+                                   max_user_code_retries,
+                                   self.ubf_context,
+                                   inputs)
 
                 # decorators can actually decorate the step function,
                 # or they can replace it altogether. This functionality
@@ -347,7 +399,8 @@ class MetaflowTask(object):
                                                self.flow,
                                                self.flow._graph,
                                                retry_count,
-                                               max_user_code_retries)
+                                               max_user_code_retries,
+                                               self.ubf_context)
 
             if join_type:
                 # Join step:
@@ -431,6 +484,9 @@ class MetaflowTask(object):
                 raise
 
         finally:
+            if self.ubf_context == UBF_CONTROL:
+                self._finalize_control_task()
+
             end = time.time() - start
 
             msg = {
@@ -444,11 +500,24 @@ class MetaflowTask(object):
             }
             logger.log(msg)
 
+            attempt_ok = str(bool(self.flow._task_ok))
+            self.metadata.register_metadata(run_id,
+                                            step_name,
+                                            task_id,
+                                            [MetaDatum(field='attempt_ok',
+                                                       value=attempt_ok,
+                                                       type='internal_attempt_status',
+                                                       tags=["attempt_id:{0}".
+                                                       format(str(retry_count))
+                                                             ])
+                                             ])
+
             output.save_metadata('task_end', {})
             output.persist(self.flow)
 
             # terminate side cars
             logger.terminate()
+            self.metadata.stop_heartbeat()
 
             # this writes a success marker indicating that the
             # "transaction" is done
